part of YIM;

class _YIMMessageManager {
  _YIMMessageManager._internal();

  static final _ = _YIMMessageManager._internal();

  factory _YIMMessageManager() {
    if (_YIMSDK.option == null) _YIMSDK.utils.yimLog("${_.runtimeType}:Please init First!");
    return _;
  }

  void clearLocalMessage(YIMEnum_ConversationType conversationType, String conversationID) {
    if (!YIM.userManager.isLogin()) return;
    switch (conversationType) {
      case YIMEnum_ConversationType.P2P:
        _YIMSDK.db.delete(
          _YIMTable.Message_P2P,
          finder: Finder(
            filter: Filter.or([
              Filter.and([Filter.equals("from", _YIMSDK.currentUser), Filter.equals("to", conversationID)]),
              Filter.and([Filter.equals("to", _YIMSDK.currentUser), Filter.equals("from", conversationID)]),
            ]),
          ),
          callback: (isSuccess) {
            if (!isSuccess) return;
            YIM.conversationManager.delete(conversationType: conversationType, conversationID: conversationID);
          },
        );
        break;
      case YIMEnum_ConversationType.Team:
        _YIMSDK.db.delete(
          _YIMTable.Message_Team,
          finder: Finder(
            filter: Filter.equals("to", conversationID),
          ),
          callback: (isSuccess) {
            if (!isSuccess) return;
            YIM.conversationManager.delete(conversationType: conversationType, conversationID: conversationID);
          },
        );
        break;
      default:
        break;
    }
  }

  void clearServerMessages({required YIMEnum_ConversationType conversationType, required String conversationID, void Function(bool isSuccess)? callback}) {
    if (!YIM.userManager.isLogin()) {
      try {
        callback?.call(false);
      } catch (e) {}
      return;
    }
    _YIMSDK.server.emit("clearServerHistory", data: {
      "conversationType": _to_YIMEnum_ConversationType(conversationType),
      "conversationID": conversationID,
    }, ack: (isSuccess, results, code, msg) {
      if (!isSuccess) {
        try {
          callback?.call(false);
        } catch (e) {}
        return;
      }
      try {
        callback?.call(true);
      } catch (e) {}
    });
  }

  void sendTextMessage({required YIMEnum_ConversationType conversationType, required String conversationID, required String content, void Function(YIMMessage message)? callback_start, void Function(bool isSuccess, YIMMessage message, int code, String msg)? callback_finish}) {
    _sendMessage(conversationType, conversationID, YIMEnum_MessageType.Text, content, callback_start, callback_finish);
  }

  void _sendMessage(YIMEnum_ConversationType conversationType, String conversationID, YIMEnum_MessageType messageType, String content, void Function(YIMMessage message)? callback_start, void Function(bool isSuccess, YIMMessage message, int code, String msg)? callback_finish) {
    final message = YIMMessage()
      ..conversationType = conversationType
      ..messageType = messageType
      ..content = content
      ..messageState = YIMEnum_MessageState.Sending
      ..from = _YIMSDK.currentUser!
      ..to = conversationID
      ..time = DateTime.now();
    // 插入一条新消息
    _YIMTable? tableName;
    switch (conversationType) {
      case YIMEnum_ConversationType.P2P:
        tableName = _YIMTable.Message_P2P;
        message.id = "message_p2p_${Ulid().toCanonical()}";
        break;
      case YIMEnum_ConversationType.Team:
        tableName = _YIMTable.Message_Team;
        message.id = "message_team_${Ulid().toCanonical()}";
        break;
      default:
        break;
    }
    try {
      callback_start?.call(message);
    } catch (e) {}
    if (!YIM.userManager.isLogin()) {
      message.messageState = YIMEnum_MessageState.Fail;
      try {
        callback_finish?.call(false, message, 0, "Please login first");
      } catch (e) {}
      return;
    }
    _YIMSDK.db.insert(
      tableName!,
      data: message,
      callback: (isSuccess) {
        if (!isSuccess) {
          message.messageState = YIMEnum_MessageState.Fail;
          try {
            callback_finish?.call(false, message, 0, "insert msg error");
          } catch (e) {}
          return;
        }
        // 创建或修改会话最后一条消息
        message.messageState = YIMEnum_MessageState.Success_UnRead;
        YIM.conversationManager._updateLastMessage(conversationType, conversationID, callback: (isSuccess) {
          if (!isSuccess) {
            message.messageState = YIMEnum_MessageState.Fail;
            try {
              callback_finish?.call(false, message, 0, "update conversation1 error");
            } catch (e) {}
            return;
          }
          try {
            YIM.listenerManager._callback_recentConversation?.call();
          } catch (e) {}
          _YIMSDK.server.emit("sendMessage", data: {
            "message": message,
          }, ack: (isSuccess, results, code, msg) {
            if (isSuccess) {
              message.messageState = YIMEnum_MessageState.Success_UnRead;
              message.time = DateTime.fromMillisecondsSinceEpoch(results!["time"]);
            } else {
              message.messageState = YIMEnum_MessageState.Fail;
            }
            // 更新新消息发送状态
            switch (conversationType) {
              case YIMEnum_ConversationType.P2P:
                _YIMSDK.db.updateOne<YIMMessage>(
                  _YIMTable.Message_P2P,
                  (oldData) => message,
                  onModel: (map) => YIMMessage.fromJson(map),
                  finder: Finder(filter: Filter.equals("id", message.id)),
                  callback: (isSuccess) {
                    if (!isSuccess) {
                      message.messageState = YIMEnum_MessageState.Fail;
                      try {
                        callback_finish?.call(false, message, 0, "update msg error");
                      } catch (e) {}
                      return;
                    }
                    // 更新会话最后一条消息
                    YIM.conversationManager._updateLastMessage(conversationType, conversationID, callback: (isSuccess) {
                      if (!isSuccess) {
                        message.messageState = YIMEnum_MessageState.Fail;
                        try {
                          callback_finish?.call(false, message, 0, "update conversation2 error");
                        } catch (e) {}
                        return;
                      }
                      try {
                        YIM.listenerManager._callback_recentConversation?.call();
                      } catch (e) {}
                      try {
                        callback_finish?.call(true, message, code, msg);
                      } catch (e) {}
                    });
                  },
                );
                break;
              case YIMEnum_ConversationType.Team:
                _YIMSDK.db.updateOne<YIMMessage>(
                  _YIMTable.Message_Team,
                  (oldData) => message,
                  onModel: (map) => YIMMessage.fromJson(map),
                  finder: Finder(filter: Filter.equals("id", message.id)),
                  callback: (isSuccess) {
                    if (!isSuccess) {
                      message.messageState = YIMEnum_MessageState.Fail;
                      try {
                        callback_finish?.call(false, message, 0, "update msg error");
                      } catch (e) {}
                      return;
                    }
                    // 更新会话最后一条消息
                    YIM.conversationManager._updateLastMessage(conversationType, conversationID, callback: (isSuccess) {
                      if (!isSuccess) {
                        message.messageState = YIMEnum_MessageState.Fail;
                        try {
                          callback_finish?.call(false, message, 0, "update conversation2 error");
                        } catch (e) {}
                        return;
                      }
                      try {
                        YIM.listenerManager._callback_recentConversation?.call();
                      } catch (e) {}
                      try {
                        callback_finish?.call(true, message, code, msg);
                      } catch (e) {}
                    });
                  },
                );
                break;
              default:
                break;
            }
          });
        });
      },
    );
  }

  void deleteLocalMessage({required YIMEnum_ConversationType conversationType, required String conversationID, required String messageID}) {
    if (!YIM.userManager.isLogin()) return;
    deleteLocalMessages(conversationType: conversationType, conversationID: conversationID, messageIDs: [messageID]);
  }

  void deleteLocalMessages({required YIMEnum_ConversationType conversationType, required String conversationID, required List<String> messageIDs}) {
    if (!YIM.userManager.isLogin()) return;
    switch (conversationType) {
      case YIMEnum_ConversationType.P2P:
        _YIMSDK.db.delete(
          _YIMTable.Message_P2P,
          finder: Finder(
            filter: Filter.and([
              Filter.inList("id", messageIDs),
              Filter.or([
                Filter.and([Filter.equals("from", _YIMSDK.currentUser), Filter.equals("to", conversationID)]),
                Filter.and([Filter.equals("to", _YIMSDK.currentUser), Filter.equals("from", conversationID)]),
              ]),
            ]),
          ),
          callback: (isSuccess) {
            if (!isSuccess) return;
            _YIMSDK.db.count(
              _YIMTable.Message_P2P,
              filter: Filter.or([
                Filter.and([Filter.equals("from", _YIMSDK.currentUser), Filter.equals("to", conversationID)]),
                Filter.and([Filter.equals("to", _YIMSDK.currentUser), Filter.equals("from", conversationID)]),
              ]),
              callback: (isSuccess, count) {
                if (!isSuccess) return;
                if (count! > 0) {
                  YIM.conversationManager._updateLastMessage(conversationType, conversationID, unRead: _YIMEnum_ConversatypeUnRead.Zero, callback: (isSuccess) {
                    if (!isSuccess) return;
                    try {
                      YIM.listenerManager._callback_recentConversation?.call();
                    } catch (e) {}
                  });
                } else {
                  YIM.conversationManager.delete(conversationType: conversationType, conversationID: conversationID);
                }
              },
            );
          },
        );
        break;
      case YIMEnum_ConversationType.Team:
        _YIMSDK.db.delete(
          _YIMTable.Message_Team,
          finder: Finder(
            filter: Filter.and([
              Filter.inList("id", messageIDs),
              Filter.equals("to", conversationID),
            ]),
          ),
          callback: (isSuccess) {
            if (!isSuccess) return;
            _YIMSDK.db.count(
              _YIMTable.Message_Team,
              filter: Filter.equals("to", conversationID),
              callback: (isSuccess, count) {
                if (!isSuccess) return;
                if (count! > 0) {
                  YIM.conversationManager._updateLastMessage(conversationType, conversationID, unRead: _YIMEnum_ConversatypeUnRead.Zero, callback: (isSuccess) {
                    if (!isSuccess) return;
                    try {
                      YIM.listenerManager._callback_recentConversation?.call();
                    } catch (e) {}
                  });
                } else {
                  YIM.conversationManager.delete(conversationType: conversationType, conversationID: conversationID);
                }
              },
            );
          },
        );
        break;
      default:
        break;
    }
  }

  void queryLocalMessages({required YIMEnum_ConversationType conversationType, required String conversationID, DateTime? time, bool onlyMine = false, int limit = 20, YIMEnum_QueryDirection queryDirection = YIMEnum_QueryDirection.Old, required void Function(List<YIMMessage>? messages) callback}) {
    if (!YIM.userManager.isLogin()) {
      try {
        callback.call(null);
      } catch (e) {}
      return;
    }
    limit = limit > 100 ? 100 : limit;
    _YIMTable? tableName;
    Filter? filter;
    switch (conversationType) {
      case YIMEnum_ConversationType.P2P:
        tableName = _YIMTable.Message_P2P;
        filter = Filter.and([
          time == null
              ? Filter.notNull("time")
              : queryDirection == YIMEnum_QueryDirection.Old
                  ? Filter.lessThan("time", time.millisecondsSinceEpoch.toString())
                  : Filter.greaterThan("time", time.millisecondsSinceEpoch.toString()),
          Filter.or([
            Filter.and([Filter.equals("from", _YIMSDK.currentUser), Filter.equals("to", conversationID)]),
            if (!onlyMine) Filter.and([Filter.equals("to", _YIMSDK.currentUser), Filter.equals("from", conversationID)]),
          ]),
        ]);
        break;
      case YIMEnum_ConversationType.Team:
        tableName = _YIMTable.Message_Team;
        filter = Filter.and([
          time == null
              ? Filter.notNull("time")
              : queryDirection == YIMEnum_QueryDirection.Old
                  ? Filter.lessThan("time", time.millisecondsSinceEpoch.toString())
                  : Filter.greaterThan("time", time.millisecondsSinceEpoch.toString()),
          Filter.equals("to", conversationID),
          if (onlyMine) Filter.equals("from", _YIMSDK.currentUser),
        ]);
        break;
    }
    _YIMSDK.db.query<YIMMessage>(
      tableName!,
      onModel: (map) => YIMMessage.fromJson(map),
      finder: Finder(
        limit: limit,
        sortOrders: [SortOrder("time", false)],
        filter: filter,
      ),
      callback: (isSuccess, results) {
        if (!isSuccess) {
          try {
            callback.call(null);
          } catch (e) {}
          return;
        }
        try {
          callback.call(results?.reversed.toList() ?? []);
        } catch (e) {}
      },
    );
  }

  void queryServerMessages({required YIMEnum_ConversationType conversationType, required String conversationID, DateTime? time, int limit = 20, YIMEnum_QueryDirection queryDirection = YIMEnum_QueryDirection.Old, required void Function(bool isSuccess, List<YIMMessage>? messages) callback}) {
    if (!YIM.userManager.isLogin()) {
      try {
        callback.call(false, null);
      } catch (e) {}
      return;
    }
    _YIMSDK.server.emit("queryMessages", data: {
      "conversationType": _to_YIMEnum_ConversationType(conversationType),
      "conversationID": conversationID,
      "time": (time ?? DateTime.now()).millisecondsSinceEpoch,
      "limit": limit,
      "queryDirection": _to_YIMEnum_QueryDirection(queryDirection),
    }, ack: (isSuccess, results, code, msg) {
      if (!isSuccess) {
        try {
          callback.call(false, null);
        } catch (e) {}
        return;
      }
      try {
        callback.call(true, (results!["messages"] as List<dynamic>).map((m) => YIMMessage.fromJson(m)).toList());
      } catch (e) {}
    });
  }

  void queryLocalMessagesByID({required YIMEnum_ConversationType conversationType, required String conversationID, required String messageID, required void Function(YIMMessage? message) callback}) {
    if (!YIM.userManager.isLogin()) {
      try {
        callback.call(null);
      } catch (e) {}
      return;
    }
    queryLocalMessagesByIDs(
      conversationType: conversationType,
      conversationID: conversationID,
      messageIDs: [messageID],
      callback: (messages) {
        try {
          callback.call(messages == null || messages.isEmpty ? null : messages.first);
        } catch (e) {}
      },
    );
  }

  void queryLocalMessagesByIDs({required YIMEnum_ConversationType conversationType, required String conversationID, required List<String> messageIDs, required void Function(List<YIMMessage>? messages) callback}) {
    if (!YIM.userManager.isLogin()) {
      try {
        callback.call(null);
      } catch (e) {}
      return;
    }
    _YIMTable? tableName;
    switch (conversationType) {
      case YIMEnum_ConversationType.P2P:
        tableName = _YIMTable.Message_P2P;
        break;
      case YIMEnum_ConversationType.Team:
        tableName = _YIMTable.Message_Team;
        break;
    }
    _YIMSDK.db.query<YIMMessage>(
      tableName!,
      onModel: (map) => YIMMessage.fromJson(map),
      finder: Finder(
        sortOrders: [SortOrder("time", true)],
        filter: Filter.inList("id", messageIDs),
      ),
      callback: (isSuccess, results) {
        if (!isSuccess) {
          try {
            callback.call(null);
          } catch (e) {}
          return;
        }
        try {
          callback.call(results ?? []);
        } catch (e) {}
      },
    );
  }

  void updateMessage({required YIMMessage message}) {
    if (!YIM.userManager.isLogin()) return;
    _YIMTable? tableName;
    switch (message.conversationType) {
      case YIMEnum_ConversationType.P2P:
        tableName = _YIMTable.Message_P2P;
        break;
      case YIMEnum_ConversationType.Team:
        tableName = _YIMTable.Message_Team;
        break;
    }
    _YIMSDK.db.updateOne<YIMMessage>(
      tableName!,
      (oldData) => message,
      onModel: (map) => YIMMessage.fromJson(map),
      finder: Finder(filter: Filter.equals("id", message.id)),
    );
  }

  void sendMessageReceipt({required YIMEnum_ConversationType conversationType, required String conversationID, void Function(bool isSuccess)? callback}) {
    if (!YIM.userManager.isLogin()) {
      try {
        callback?.call(false);
      } catch (e) {}
      return;
    }
    if (conversationType != YIMEnum_ConversationType.P2P) return;
    _YIMSDK.server.emit("sendMessageReceipt", data: {
      "conversationType": _to_YIMEnum_ConversationType(conversationType),
      "conversationID": conversationID,
    }, ack: (isSuccess, results, code, msg) {
      if (!isSuccess) {
        try {
          callback?.call(false);
        } catch (e) {}
        return;
      }
      _markConversationRead(
        conversationID,
        callback: (isSuccess) {
          if (!isSuccess) {
            try {
              callback?.call(false);
            } catch (e) {}
            return;
          }
          try {
            callback?.call(true);
          } catch (e) {}
        },
      );
    });
  }

  void sendMessageRevoke({required YIMEnum_ConversationType conversationType, required String conversationID, required String messageID, void Function(bool isSuccess)? callback}) {
    if (!YIM.userManager.isLogin()) {
      try {
        callback?.call(false);
      } catch (e) {}
      return;
    }
    _YIMSDK.server.emit("sendMessageRevoke", data: {
      "conversationType": _to_YIMEnum_ConversationType(conversationType),
      "conversationID": conversationID,
      "messageID": messageID,
    }, ack: (isSuccess, results, code, msg) {
      if (!isSuccess) {
        try {
          callback?.call(false);
        } catch (e) {}
        return;
      }
      _markConversationRevoke(
        conversationType,
        messageID,
        callback: (isSuccess) {
          if (!isSuccess) {
            try {
              callback?.call(false);
            } catch (e) {}
            return;
          }
          YIM.conversationManager._updateLastMessage(
            conversationType,
            conversationID,
            // unRead: _YIMEnum_ConversatypeUnRead.Minus,
            callback: (isSuccess) {
              if (!isSuccess) {
                try {
                  callback?.call(false);
                } catch (e) {}
                return;
              }
              try {
                YIM.listenerManager._callback_recentConversation?.call();
              } catch (e) {}
              try {
                callback?.call(true);
              } catch (e) {}
            },
          );
        },
      );
    });
  }

  void _markConversationRead(String conversationID, {void Function(bool isSuccess)? callback}) {
    _YIMSDK.db.updateAll(
      _YIMTable.Message_P2P,
      () => {"messageState": _to_YIMEnum_MessageState(YIMEnum_MessageState.Success_Read)},
      finder: Finder(
        filter: Filter.or([
          Filter.and([Filter.equals("from", _YIMSDK.currentUser), Filter.equals("to", conversationID)]),
          Filter.and([Filter.equals("to", _YIMSDK.currentUser), Filter.equals("from", conversationID)]),
        ]),
      ),
      callback: (isSuccess) {
        if (!isSuccess) {
          try {
            callback?.call(false);
          } catch (e) {}
          return;
        }
        try {
          callback?.call(true);
        } catch (e) {}
      },
    );
  }

  void _markConversationRevoke(YIMEnum_ConversationType conversationType, String messageID, {void Function(bool isSuccess)? callback}) {
    _YIMTable? tableName;
    switch (conversationType) {
      case YIMEnum_ConversationType.P2P:
        tableName = _YIMTable.Message_P2P;
        break;
      case YIMEnum_ConversationType.Team:
        tableName = _YIMTable.Message_Team;
        break;
    }
    _YIMSDK.db.updateAll(
      tableName!,
      () => {"isRevoke": 1},
      finder: Finder(
        filter: Filter.equals("id", messageID),
      ),
      callback: (isSuccess) {
        if (!isSuccess) {
          try {
            callback?.call(false);
          } catch (e) {}
          return;
        }
        try {
          callback?.call(true);
        } catch (e) {}
      },
    );
  }
}
