part of YIM;

// Conversation：lastMessage unRead
// Message_P2P：id content from to conversationType messageType messageState time localExt isRevoke
// Message_Team：id content from to conversationType messageType messageState time localExt isRevoke
enum _YIMTable { Conversation, Message_P2P, Message_Team }

class _YIMDB {
  Map<String, Database> _databaseClients = {};

  Future<Database?> _getDatabaseClient(_YIMTable tableName) async {
    if (_YIMSDK.option == null) {
      _YIMSDK.utils.yimLog("DB: Please init First!");
      return null;
    }
    if (!YIM.userManager.isLogin()) {
      _YIMSDK.utils.yimLog("DB: Please login first!");
      return null;
    }
    final dbName = tableName.name;
    if (!_databaseClients.containsKey(dbName)) {
      final path = "/yim/${_YIMSDK.option!.appKey}/${_YIMSDK.currentUser}/${dbName}.db";
      _databaseClients[dbName] = kIsWeb ? await databaseFactoryWeb.openDatabase(path) : await databaseFactoryIo.openDatabase("${(await getApplicationDocumentsDirectory()).path}${path}");
    }
    return _databaseClients[dbName]!;
  }

  StoreRef<int, Map<String, Object?>> _getDB() => intMapStoreFactory.store(null);

  void count(_YIMTable tableName, {Filter? filter, required void Function(bool isSuccess, int? count) callback}) {
    _getDatabaseClient(tableName).then((databaseClient) {
      if (databaseClient == null) {
        callback(false, null);
        return;
      }
      try {
        _getDB().count(databaseClient, filter: filter).then((count) {
          callback(true, count);
        });
      } catch (e) {
        _YIMSDK.utils.yimLog("DB_c: ${e}");
        callback(false, null);
      }
    });
  }

  void query<T extends Object>(_YIMTable tableName, {Finder? finder, required onModel(map)?, required void Function(bool isSuccess, List<T>? results) callback}) {
    _getDatabaseClient(tableName).then((databaseClient) {
      if (databaseClient == null) {
        callback(false, null);
        return;
      }
      try {
        _getDB().find(databaseClient, finder: finder).then((List<dynamic> db_results) {
          if (db_results.isEmpty) {
            callback(true, []);
            return;
          }
          if (onModel == null)
            callback(true, db_results as List<T>);
          else
            callback(true, db_results.map<T>((e) => onModel(e.value) as T).toList());
        });
      } catch (e) {
        _YIMSDK.utils.yimLog("DB_q: ${e}");
        callback(false, null);
      }
    });
  }

  void insert(_YIMTable tableName, {Object? data, List<Object>? datas, void Function(bool isSuccess)? callback}) async {
    if (data == null && datas == null && datas!.isEmpty) {
      _YIMSDK.utils.yimLog("DB_i: No Data");
      callback?.call(false);
      return;
    }
    final databaseClient = await _getDatabaseClient(tableName);
    if (databaseClient == null) {
      callback?.call(false);
      return;
    }
    try {
      _getDB().addAll(databaseClient, (datas ?? [data!]).map<Map<String, Object?>>((e) => json.decode(_YIMSDK.utils.toJson(e))).toList());
      if (callback != null) Future.delayed(Duration(milliseconds: _YIMSDK.config.dbDelayTime), () => callback.call(true));
    } catch (e) {
      _YIMSDK.utils.yimLog("DB_i: ${e}");
      callback?.call(false);
    }
  }

  // 用于修改查询结果仅存在1条的数据
  void updateOne<T extends Object>(_YIMTable tableName, T Function(T? oldData) onQueryFinish, {Finder? finder, required onModel(map)?, void Function(bool isSuccess)? callback}) async {
    final databaseClient = await _getDatabaseClient(tableName);
    if (databaseClient == null) {
      callback?.call(false);
      return;
    }
    try {
      query<T>(tableName, finder: finder, onModel: onModel, callback: (isSuccess, results) {
        if (!isSuccess) {
          callback?.call(false);
          return;
        }
        if (results!.isNotEmpty) _getDB().update(databaseClient, json.decode(_YIMSDK.utils.toJson(onQueryFinish(results.first))), finder: finder);
        if (callback != null) Future.delayed(Duration(milliseconds: _YIMSDK.config.dbDelayTime), () => callback.call(true));
      });
    } catch (e) {
      _YIMSDK.utils.yimLog("DB_uo: ${e}");
      callback?.call(false);
    }
  }

  // 用于修改查询结果大于等于0条的数据，批量修改指定字段
  void updateAll(_YIMTable tableName, Map<String, dynamic> Function() onQueryFinish, {Finder? finder, void Function(bool isSuccess)? callback}) async {
    final databaseClient = await _getDatabaseClient(tableName);
    if (databaseClient == null) {
      callback?.call(false);
      return;
    }
    try {
      _getDB().update(databaseClient, json.decode(_YIMSDK.utils.toJson(onQueryFinish())), finder: finder);
      if (callback != null) Future.delayed(Duration(milliseconds: _YIMSDK.config.dbDelayTime), () => callback.call(true));
    } catch (e) {
      _YIMSDK.utils.yimLog("DB_ua: ${e}");
      callback?.call(false);
    }
  }

  Future<void> delete(_YIMTable tableName, {Finder? finder, void Function(bool isSuccess)? callback}) async {
    final databaseClient = await _getDatabaseClient(tableName);
    if (databaseClient == null) {
      callback?.call(false);
      return;
    }
    try {
      _getDB().delete(databaseClient, finder: finder);
      if (callback != null) Future.delayed(Duration(milliseconds: _YIMSDK.config.dbDelayTime), () => callback.call(true));
    } catch (e) {
      _YIMSDK.utils.yimLog("DB_d:${e}");
      callback?.call(false);
    }
  }

  void updateOrInsert<T extends Object>(_YIMTable tableName, T Function(T? data) onQueryFinish, {Finder? finder, required onModel(map)?, void Function(bool isSuccess)? callback}) {
    try {
      query<T>(tableName, finder: finder, onModel: onModel, callback: (isSuccess, results) {
        try {
          if (!isSuccess) {
            callback?.call(false);
            return;
          }
          if (results!.isEmpty)
            insert(tableName, data: onQueryFinish(null), callback: callback);
          else
            updateOne<T>(tableName, (oldData) => onQueryFinish(oldData), finder: finder, onModel: onModel, callback: callback);
        } catch (e) {
          _YIMSDK.utils.yimLog("DB_uoi:${e}");
          callback?.call(false);
        }
      });
    } catch (e) {
      _YIMSDK.utils.yimLog("DB_uoi:${e}");
      callback?.call(false);
    }
  }

  void dispose() {
    _databaseClients.values.forEach((element) => element.close());
    _databaseClients.clear();
  }
}
