part of YIM;

class _YIMServer {
  void Function(_YIMResult<YIMEnum_LoginResult> result)? callback_login;
  Socket? socket;
  Timer? timer_checkConnect;
  Timer? timer_login;
  var isLoginDone = false;
  String? currentToken;

  void init(String user, String token, void Function(_YIMResult<YIMEnum_LoginResult> result)? callback) {
    isLoginDone = false;
    this.callback_login = callback;
    _YIMSDK.currentUser = user;
    currentToken = token;
    dispose(false);
    connect(false);
    timer_login = Timer.periodic(Duration(milliseconds: _YIMSDK.config.socketServerConnectTimeout), (timer) {
      onServer("loginFailed", {"msg": "time out"});
    });
  }

  void connect(bool isReconnect) {
    final params = {
      "accid": _YIMSDK.currentUser,
      "token": currentToken,
      "appKey": _YIMSDK.option?.appKey ?? "",
      "device": kIsWeb ? "web" : Platform.operatingSystem,
      "isReconnect": isReconnect ? "1" : "0",
    };
    final option = OptionBuilder();
    option.setTimeout(_YIMSDK.config.socketServerConnectTimeout);
    option.setTransports(['websocket']);
    option.disableAutoConnect();
    option.disableReconnection();
    option.setQuery(params);
    socket = io(_YIMSDK.config.server, option.build());
    socket!.onConnect((data) => _YIMSDK.utils.yimLog("onConnect"));
    socket!.onReconnect((data) => _YIMSDK.utils.yimLog("onReconnect"));
    socket!.onConnecting((data) => _YIMSDK.utils.yimLog("onConnecting"));
    socket!.onConnectTimeout((data) => _YIMSDK.utils.yimLog("onConnectTimeout"));
    socket!.onConnectError((data) => _YIMSDK.utils.yimLog("onConnectError"));
    socket!.onDisconnect((data) => _YIMSDK.utils.yimLog("onDisconnect"));
    socket!.on('server', (data_server) => onServer(data_server["event"], data_server["data"]));
    socket!.connect();
  }

  void emit(String event, {Map<String, Object?>? data, void Function(bool isSuccess, Map<String, dynamic>? results, int code, String msg)? ack, int retryTime = 0}) {
    try {
      if (retryTime > _YIMSDK.config.maxSocketServerRetryTimes) {
        try {
          ack?.call(false, null, 0, "[$event] time out");
        } catch (e) {}
        return;
      }
      if (!YIM.userManager.isLogin()) {
        try {
          ack?.call(false, null, 0, "Please login first!");
        } catch (e) {}
        return;
      }
      var isFinish = false;
      final timer = Timer.periodic(Duration(milliseconds: _YIMSDK.config.socketServerRetryTimeout), (timer) {
        if (!isFinish) {
          retryTime++;
          if (retryTime > _YIMSDK.config.maxSocketServerRetryTimes) {
            try {
              ack?.call(false, null, 0, "[$event] time out");
            } catch (e) {}
            timer.cancel();
            return;
          }
          _YIMSDK.utils.yimLog("[$event] emit time out,will retry,current times:$retryTime");
          emit(event, data: data, ack: ack, retryTime: retryTime);
        }
        timer.cancel();
      });
      _YIMSDK.utils.yimLog("[emit_request]: \nevent: $event\ndata: ${_YIMSDK.utils.toJson(data ?? {})}");
      socket?.emitWithAck("client", {
        "event": event,
        "data": data,
      }, ack: (data) {
        try {
          _YIMSDK.utils.yimLog("[emit_response]: \nevent: $event\ndata: ${_YIMSDK.utils.toJson(data ?? {})}");
          isFinish = true;
          timer.cancel();
          Map<String, dynamic>? results = data;
          if (results == null) {
            try {
              ack?.call(false, null, 0, "null response");
            } catch (e) {}
            return;
          }
          final code = results["code"] ?? 0;
          final isSuccess = (code == 200);
          final msg = results["msg"] ?? "";
          results.remove("code");
          results.remove("msg");
          try {
            ack?.call(isSuccess, results["data"], code, msg);
          } catch (e) {}
        } catch (e) {
          try {
            ack?.call(false, null, 0, "Exception: ${e.toString()}");
          } catch (e) {}
        }
      });
    } catch (e) {
      try {
        ack?.call(false, null, 0, "Exception: ${e.toString()}");
      } catch (e) {}
    }
  }

  void dispose(bool isReconnect) {
    if (isReconnect) {
      socket?.clearListeners();
      socket?.dispose();
      socket = null;
    } else {
      // currentToken = null;
      timer_checkConnect?.cancel();
      timer_checkConnect = null;
      timer_login?.cancel();
      timer_login = null;
      socket?.clearListeners();
      socket?.dispose();
      socket = null;
    }
  }

  List<YIMMessage> cacheMessages = [];
  var lastMessageTime = 0;
  Timer? timer_message;

  Future<void> onServer(String event, Map<String, dynamic> data) async {
    void doMessages(List<YIMMessage> messages_do, bool isNewMessage, {void Function(bool isSuccess)? callback}) {
      List<YIMMessage> messages = [];
      messages_do.forEach((message_do) {
        if (!messages.contains(message_do)) messages.add(message_do);
      });
      for (var i = 0; i < messages.length; i++) {
        final message = messages[i];
        switch (message.conversationType) {
          case YIMEnum_ConversationType.P2P:
            _YIMSDK.db.count(
              _YIMTable.Message_P2P,
              filter: Filter.equals("id", message.id),
              callback: (isSuccess, count) {
                if (!isSuccess) {
                  if (i == messages.length - 1) callback?.call(false);
                  return;
                }
                if (count! <= 0) {
                  if (i == messages.length - 1) callback?.call(false);
                  return;
                }
                _YIMSDK.db.insert(
                  _YIMTable.Message_P2P,
                  data: message,
                  callback: (isSuccess) {
                    if (!isSuccess) {
                      if (i == messages.length - 1) callback?.call(false);
                      return;
                    }
                    YIM.conversationManager._updateLastMessage(
                      message.conversationType,
                      message.from == _YIMSDK.currentUser ? message.to : message.from,
                      unRead: _YIMEnum_ConversatypeUnRead.Plus,
                      callback: (isSuccess) {
                        if (!isSuccess) {
                          if (i == messages.length - 1) callback?.call(false);
                          return;
                        }
                        try {
                          YIM.listenerManager._callback_recentConversation?.call();
                        } catch (e) {}
                        if (isNewMessage)
                          try {
                            YIM.listenerManager._callback_message?.call(messages);
                          } catch (e) {}
                        if (i == messages.length - 1) callback?.call(true);
                      },
                    );
                  },
                );
              },
            );
            break;
          case YIMEnum_ConversationType.Team:
            _YIMSDK.db.count(
              _YIMTable.Message_Team,
              filter: Filter.equals("id", message.id),
              callback: (isSuccess, count) {
                if (!isSuccess) {
                  if (i == messages.length - 1) callback?.call(false);
                  return;
                }
                if (count! <= 0) {
                  if (i == messages.length - 1) callback?.call(false);
                  return;
                }
                _YIMSDK.db.insert(
                  _YIMTable.Message_Team,
                  data: message,
                  callback: (isSuccess) {
                    if (!isSuccess) {
                      if (i == messages.length - 1) callback?.call(false);
                      return;
                    }
                    YIM.conversationManager._updateLastMessage(
                      message.conversationType,
                      message.to,
                      unRead: _YIMEnum_ConversatypeUnRead.Plus,
                      callback: (isSuccess) {
                        if (!isSuccess) {
                          if (i == messages.length - 1) callback?.call(false);
                          return;
                        }
                        try {
                          YIM.listenerManager._callback_recentConversation?.call();
                        } catch (e) {}
                        if (isNewMessage)
                          try {
                            YIM.listenerManager._callback_message?.call(messages);
                          } catch (e) {}
                        if (i == messages.length - 1) callback?.call(true);
                      },
                    );
                  },
                );
              },
            );
            break;
          default:
            break;
        }
      }
    }

    switch (event) {
      case "loginFailed":
        if (isLoginDone) return;
        isLoginDone = true;
        YIM.userManager.logout();
        try {
          callback_login?.call(_YIMResult<YIMEnum_LoginResult>(code: YIMEnum_LoginResult.Failed, msg: data["msg"]));
        } catch (e) {}
        callback_login = null;
        break;
      case "loginSuccess":
        if (isLoginDone) return;
        isLoginDone = true;
        timer_checkConnect?.cancel();
        timer_checkConnect = null;
        timer_login?.cancel();
        timer_login = null;
        timer_checkConnect = Timer.periodic(Duration(seconds: 5), (timer) {
          if (!YIM.userManager.isLogin()) {
            try {
              YIM.listenerManager._callback_status_online?.call(_YIMResult<YIMEnum_OnlineState>(code: YIMEnum_OnlineState.Reconnect, msg: "Reconnect"));
            } catch (e) {}
            dispose(true);
            connect(true);
          } else
            _YIMSDK.server.emit("refreshOnline");
        });
        //清理旧消息
        _YIMSDK.db.count(
          _YIMTable.Message_Team,
          callback: (isSuccess, count) {
            if (!isSuccess) return;
            if (count! > _YIMSDK.config.maxTeamMessageCount)
              _YIMSDK.db.delete(
                _YIMTable.Message_Team,
                finder: Finder(
                  limit: _YIMSDK.config.maxTeamMessageCount ~/ 2,
                  sortOrders: [SortOrder("time", false)],
                ),
              );
          },
        );
        //同步云端会话
        emit("queryConversations", ack: (isSuccess, results, code, msg) {
          if (isSuccess) {
            List<Map<String, dynamic>> conversations_server = (json.decode(results?["conversations"]?.toString() ?? "[]") as List<dynamic>).map((m) => m as Map<String, dynamic>).toList();
            if (conversations_server.isEmpty) return;
            YIM.conversationManager.queryRecentConversation(
              callback: (conversations) {
                //过滤本地已经存在的云端会话
                conversations?.forEach((conversation) {
                  switch (conversation.lastMessage!.conversationType) {
                    case YIMEnum_ConversationType.P2P:
                      final conversationID = conversation.lastMessage!.from == _YIMSDK.currentUser ? conversation.lastMessage!.to : conversation.lastMessage!.from;
                      for (var i = conversations_server.length - 1; i >= 0; i--) {
                        if (conversations_server[i]["conversationID"] == conversationID) conversations_server.removeAt(i);
                      }
                      break;
                    case YIMEnum_ConversationType.Team:
                      final conversationID = conversation.lastMessage!.to;
                      for (var i = conversations_server.length - 1; i >= 0; i--) {
                        if (conversations_server[i]["conversationID"] == conversationID) conversations_server.removeAt(i);
                      }
                      break;
                    default:
                      break;
                  }
                });
                //将本地没有的云端会话，把消息同步下来
                for (var i = 0; i < conversations_server.length; i++) {
                  final conversation = conversations_server[i];
                  YIM.messageManager.queryServerMessages(
                    conversationType: _get_YIMEnum_ConversationType(conversation["conversationType"].toString()),
                    conversationID: conversation["conversationID"].toString(),
                    limit: _YIMSDK.config.roamMessageSize,
                    callback: (isSuccess, messages) {
                      if (isSuccess) {
                        doMessages(messages!, false, callback: (isSuccess) {
                          if (!isSuccess) return;
                          Filter? filter;
                          switch (_get_YIMEnum_ConversationType(conversation["conversationType"].toString())) {
                            case YIMEnum_ConversationType.P2P:
                              filter = Filter.and([
                                Filter.or([
                                  Filter.and([Filter.equals("lastMessage.from", _YIMSDK.currentUser), Filter.equals("lastMessage.to", conversation["conversationID"])]),
                                  Filter.and([Filter.equals("lastMessage.to", _YIMSDK.currentUser), Filter.equals("lastMessage.from", conversation["conversationID"])]),
                                ]),
                              ]);
                              break;
                            case YIMEnum_ConversationType.Team:
                              filter = Filter.and([
                                Filter.equals("lastMessage.to", conversation["conversationID"]),
                              ]);
                              break;
                          }
                          _YIMSDK.db.updateOne<YIMConversation>(
                            _YIMTable.Conversation,
                            (oldData) => oldData!..unRead = int.parse(conversation["unReadCount"].toString()),
                            onModel: (map) => YIMConversation.fromJson(map),
                            finder: Finder(filter: filter),
                            callback: (isSuccess) {
                              if (!isSuccess) return;
                              try {
                                YIM.listenerManager._callback_recentConversation?.call();
                              } catch (e) {}
                            },
                          );
                        });
                      }
                    },
                  );
                }
              },
            );
          }
        });
        try {
          callback_login?.call(_YIMResult<YIMEnum_LoginResult>(code: YIMEnum_LoginResult.Success, msg: "success"));
        } catch (e) {}
        callback_login = null;
        break;
      case "kick":
        YIM.userManager.logout();
        try {
          YIM.listenerManager._callback_status_online?.call(_YIMResult<YIMEnum_OnlineState>(code: YIMEnum_OnlineState.Kick, msg: "Kicked by ${data["device"]} device"));
        } catch (e) {}
        break;
      case "message":
        final messages = (data["messages"] as List<dynamic>).map((element) => YIMMessage.fromJson(element)).toList();
        final curTime = DateTime.now().millisecondsSinceEpoch;
        if (curTime - lastMessageTime < _YIMSDK.config.messageSpace) {
          messages.forEach((message) {
            if (!cacheMessages.contains(message)) cacheMessages.add(message);
          });
          timer_message?.cancel();
          timer_message = Timer.periodic(Duration(milliseconds: 500), (timer) {
            onServer(event, data);
            timer.cancel();
          });
        } else {
          doMessages(cacheMessages + messages, true);
          lastMessageTime = curTime;
          cacheMessages.clear();
        }
        break;
      case "broadcast":
        try {
          YIM.listenerManager._callback_broadcast?.call(_YIMBroadcastMessage.fromJson(data["message"]));
        } catch (e) {}
        break;
      case "notification":
        try {
          YIM.listenerManager._callback_notification?.call(_YIMNotificationMessage.fromJson(data["message"]));
        } catch (e) {}
        break;
      case "revoke":
        (data["messages"] as List<dynamic>).forEach((element) {
          final message = YIMRevokeMessage.fromJson(element);
          YIM.messageManager._markConversationRevoke(
            message.conversationType,
            message.messageID,
            callback: (isSuccess) {
              if (!isSuccess) return;
              YIM.conversationManager._updateLastMessage(
                message.conversationType,
                message.conversationID,
                // unRead: _YIMEnum_ConversatypeUnRead.Minus,
                callback: (isSuccess) {
                  if (!isSuccess) return;
                  try {
                    YIM.listenerManager._callback_recentConversation?.call();
                  } catch (e) {}
                  try {
                    YIM.listenerManager._callback_revoke?.call(message);
                  } catch (e) {}
                },
              );
            },
          );
        });
        break;
      case "receipt":
        List<String> conversationIDs = (data["conversationIDs"] as List).map((element) => element.toString()).toList();
        conversationIDs.forEach((conversationID) {
          YIM.messageManager._markConversationRead(
            conversationID,
            callback: (isSuccess) {
              if (!isSuccess) return;
              YIM.conversationManager._updateLastMessage(
                YIMEnum_ConversationType.P2P,
                conversationID,
                callback: (isSuccess) {
                  if (!isSuccess) return;
                  try {
                    YIM.listenerManager._callback_recentConversation?.call();
                  } catch (e) {}
                  try {
                    YIM.listenerManager._callback_status_msg_receipt?.call(conversationID);
                  } catch (e) {}
                },
              );
            },
          );
        });
        break;
    }
  }
}
